__all__ = [
    "__title__", "__summary__", "__uri__", "__version__", "__author__",
    "__email__", "__license__", "__copyright__",
]

__title__ = "vivarium"
__summary__ = "vivarium is a microsimulation framework built on top of the standard scientific python stack."
__uri__ = "https://github.com/ihmeuw/vivarium"

__version__ = "0.8.2"

__author__ = "The vivarium developers"
__email__ = "collijk@uw.edu"

__license__ = "GNU GPLv3"
__copyright__ = f"Copyright 2016-2018 {__author__}"
